###############################################
##                                           ##
##    Teilskript 3: Artverbreitungsmodelle   ##
##                                           ##
###############################################


#############################################################
# falls Session unterbrochen wurde: risktool-settings.r erneut laden.
#source("./code/risktool-settings.r")


###################################################################
#       Vorbereitung der Daten fuer die Modellierung              #
###################################################################

#############################################################
# Einlesen der Umweltvariablen

# Definition der zu verwendenden Umweltvariablen, basierend auf der
# Variablenselektion in "02_variable_selection.r"
envir.names <- c("bio_1", "bio_6", "bio_10", "bio_11")

# Einlesen der Umweltdaten fuer aktuelle Klimabedingungen in ein "RasterStack"
envir.baseline <- stack(paste0(path.envir.baseline.cropped, envir.names, ".tif"))

# Einlesen der Umweltdaten fuer zukuenftige Klimabedingungen
envir.future <- stack(paste0(path.envir.future.cropped, envir.names, ".tif"))

#############################################################
# Einlesen der Verbreitungsdaten
# mit read.csv"Dateiname", header:Spaltenueberschriften vorhanden?, 
# sep: Trennung zwischen den Spalten (";" oder "," oder "\t" (Tabulator)), dec: Dezimaltrennzeichen (am besten "."))
# Hinweis: Objektnamen (wie hier presences) sind frei waehlbar, sollten jedoch nicht den gleichen Namen wie R-Funktionen haben
presences <- read.csv(paste0(path.speciesdata, "Ae_japonicus_Japan.csv"), header=TRUE, sep=";", dec=".")


###################
# Optional: Zusammenfassung der Daten (Ueberblick)
summary(presences)
# Struktur des Objekts
str(presences)
# Zeilen 1 bis 6 der Datentabelle ausgeben
head(presences)

###################
# Extraktion der Spalten mit den Koordinaten der Vorkommenspunkte (hier: Spalte longitude und Spalte latitude)
presences <- presences[,c("longitude", "latitude")]

# Entfernen von fehlenden Werten ("NA" = not available) aus dem Datensatz; entfernt die gesamte Zeile im Datensatz
# (Beispieldatensatz enthaelt keine fehlenden Daten, daher ohne Auswirkungen)
presences <- na.omit(presences)

#############################################################
# Anpassung der Vorkommensdaten an das Raster der Umweltdaten
# Ziel: maximal ein Vorkommenspunkt pro Rasterzelle

###################
# Umwandlung der Koordinaten von "presences" in ein Raster
# mit rasterize unter Verwendung des Rasters von envir.baseline
# Ergebnis: Raster-Objekt mit Information (Daten) von envir.baseline
# in den Zellen mit den Vorkommen, alle anderen sind NA
presences.raster <- rasterize(presences, envir.baseline[[1]])

# Alle Zellen, die nicht NA sind (also die mit Vorkommen), werden mit 1 (Vorkommen) ersetzt
presences.raster[!is.na(presences.raster)] <- 1

###################
# Umwandlung der Raster-Vorkommensdaten in Punktdaten (SpatialPointsDataFrame)
# zur Verwendung in der Modellierung
presences.points <- as(presences.raster, "SpatialPointsDataFrame")


###################################################################
#                  Modellierung mit biomod2                       #
###################################################################

#############################################################
# 1. Formatierung der Daten fuer die Verwendung in biomod2
# Bei Vorkommen/Nicht-Vorkommen Daten (presence-absence) muessen keine Pseudo-Absenzen gebildet werden.
# Bei nur Vorkommen Daten (presence-only) werden die Anzahl und die Methode zur Wahl der Pseud-Absenzen festgelegt.
# PA.strategy: "random", "sre", "disk" oder "user.defined"
myBiomodData <- BIOMOD_FormatingData(resp.var = presences.points,  # Response variable (abhaengige V. (Vorkommen))
                                     expl.var = envir.baseline,    # Explanatory variable (erklaerende V. (Umwelt))
                                     resp.name = speciesname,      # Response name (Name der Art)
                                     PA.nb.rep = 1,                # Anzahl an Pseudo-Absence Selektionen (Wie oft sollen PAs gewaehlt werden)
                                     PA.nb.absences = 580,         # Anzahl an Pseudo-Absences (Wieviele PAs sollen gewaehlt werden)
                                     PA.strategy = "random")       # Auswahl-Strategie (Wie sollen PAs gewawhlt werden)  

# Abspeichern der formatierten Daten wird empfohlen, da je nach Datensatzgroesse die Formatierung sehr zeitintensiv sein kann
# Funktion: save(Objektname, Dateiname mit Pfadangabe)
# kann mit load() wieder eingelesen werden
save(myBiomodData, file=paste0(path.data, speciesname, "formatted_for_biomod.RData"))

#############################################################
# 2. Definition der Modelloptionen
# Standardeinstellungen: es werden keine Aenderungen an den Standardeinstellungen vorgenommen
myBiomodOption <- BIOMOD_ModelingOptions()

# Bei Verwendung von MAXENT muessen die Einstellungen angepasst werden!; Aenderungen der Einstellungen sind kommentiert,
# alles andere sind Standardeinstellungen von MAXENT
myBiomodOption <- BIOMOD_ModelingOptions(
  MAXENT.Phillips = list( path_to_maxent.jar = getwd(),  # Pfadangabe zu maxent.jar (ausfuehrbare Java-Datei) - muss individuell angepasst werden
                 memory_allocated = 512,        # Erhoehung des verwendeten Speichers (Standard bei Java: 64MB)
                 maximumiterations = 200,
                 visible = FALSE,
                 linear = TRUE,
                 quadratic = TRUE,
                 product = TRUE,
                 threshold = TRUE,
                 hinge = TRUE,
                 lq2lqptthreshold = 80,
                 l2lqthreshold = 10,
                 hingethreshold = 15,
                 beta_threshold = -1,
                 beta_categorical = -1,
                 beta_lqp = -1,
                 beta_hinge = -1,
                 defaultprevalence = 0.5))

#############################################################
# 3. Modellierung mit biomod2
mymodels <- BIOMOD_Modeling( myBiomodData,                             # Formatierte Daten
                                     models = c("MAXENT", "GLM", "GBM", "RF"), # zu verwendende Modellalgorithmen (10 verfuegbar)
                                     models.options = myBiomodOption,          # Modelloptionen (evtl. angepasst)
                                     NbRunEval=1,                              # Anzahl an Durchlaeufen je Modellalgorithmus
                                     DataSplit=70,                             # Validierung der Daten (70% Training, 30% Test)
                                     Prevalence=NULL,
                                     VarImport=3,                              # Anzahl an Permutationen zur Ermittlung der Wichtigkeit der einzelnen Variablen
                                     models.eval.meth = c("ROC", "TSS", "KAPPA"), # Modellguetemasse
                                     SaveObj = TRUE,                           # Sicherung aller Ergebnisse und erstellten Objekte
                                     do.full.models=FALSE,                     # Verwendung des gesamten Datensatzes falls TRUE
                                     modeling.id=speciesname,                  # Name fuer den Modellierungsdurchlauf (individuell)
                                     rescal.all.models = FALSE)                # Skalierung der Modellvorhersage mit einem GLM falls TRUE

# Sicherung des gesamten Arbeitsbereiches (work space) mit save.image als R.Data
# (kann mit load("XY.RData") wieder in R eingelesen werden)
save.image(paste0(path.data, "workspace-backup_", speciesname, ".RData"))

# Zusammenfassung der Modellierung
mymodels

# Modellevaluierung
get_evaluations(mymodels)

# Ausgabe der Wichtigkeit der einzelnen Variablen
get_variables_importance(mymodels)

#############################################################
# 4. Projektion der Modellierungsergebnisse unter aktuellen Umweltbedingungen fuer die einzelnen Modellalgorithmen
# Zur naeheren Erlaeuterung siehe auch: http://www.inside-r.org/packages/cran/biomod2/docs/BIOMOD_Projection
myproj.baseline <- BIOMOD_Projection(modeling.output = mymodels,   # Modellierungsergebnis (erstelltes Objekt)
                                  new.env = envir.baseline,        # Set an erklaerenden (Umwelt-)Variablen
                                  proj.name = "baseline",          # Name fuer die Projektion (frei waehlbar;legt einen neuen Ordner mit diesem Namen an)
                                  selected.models = 'all',         # Auswahl an Modellen, die verwendet werden sollen (all: alle Modelle werden verwendet)
                                  binary.meth = 'ROC',             # Methode zur Umrechnung der Projektion in 0/1 Daten (binaere Daten)
                                  compress = F,                    # Komprimierungsmethode beim Abspeichern von Objekten
                                  build.clamping.mask = F,         # Wenn TRUE: Identifiziert Regionen, in denen die Projektionen unsicher sind
                                  output.format = '.img')          # Objekt kann als .RData, .grd oder .img abgespeichert werden

# Zusammenfassung der Gegenwarts-Projektion
myproj.baseline

# Darstellung der Gegenwarts-Projektionen der einzelnen Modell-Algorithmen
plot(myproj.baseline)

# Ausgabe der Projektionen als Multi-layer Geo-Tiff
writeRaster(get_predictions(myproj.baseline), paste0(path.output, speciesname, "_all_models_baseline"), format="GTiff")


#############################################################
# 5. Ensemble-Modellierung Gegenwart (Berechnung eines Ensembles aus den bereits erstellten Einzel-Modellen)
# Zur naeheren Erlaeuterung siehe auch: http://www.inside-r.org/packages/cran/biomod2/docs/BIOMOD_EnsembleModeling
mymodels.ensemble <- BIOMOD_EnsembleModeling(modeling.output = mymodels,             # Modellierungsergebnis (erstelltes Objekt)
                                      chosen.models = 'all',                     # Wahl der Modellalgorithmen fuer Ensemble-Berechnung
                                      em.by='all',                               # Methode der Kombination der Modellalgorithmen
                                      eval.metric = c('ROC'),                    # Evaluationsmethode (AUC/ROC, Kappa, TSS)
                                      prob.mean = T,                             # Mittlere Wahrscheinlichkeit ueber alle Vorhersagen
                                      prob.cv = T,                               # Variationskoeffizient ueber alle Vorhersagen
                                      prob.ci = T,                               # Konfidenzintervall um "prob.mean"
                                      prob.ci.alpha = 0.05,                      # Signifikanzlevel fuer Konfidenzintervall
                                      prob.median = T,                           # Median der Wahrscheinlichkeiten
                                      committee.averaging = T,                   # Mittelwert der binaeren Vorhersage
                                      prob.mean.weight = T,                      # Gewichtete Summe der Wahrscheinlichkeiten
                                      prob.mean.weight.decay = 'proportional' )  # Relative Wichtigkeit der Gewichtung

# Ausgabe einer Zusammenfassung der Ensemble-Modellierung Gegenwart
mymodels.ensemble

# Ausgabe der Modellevaluierung
get_evaluations(mymodels.ensemble)

#############################################################
# 6. Projektion der Modellierungsergebnisse unter aktuellen Umweltbedingungen fuer die Ensemble-Modellierung

# Projektion des Ensemble-Modells auf die Baseline-Klimadaten zur Darstellung als Karte
myproj.baseline.ensemble <- BIOMOD_EnsembleForecasting(EM.output = mymodels.ensemble, # Ergebnis der Ensembleberechnung fuer die Gegenwart
                                         projection.output = myproj.baseline)         # Projektion der Einzel-Modelle fuer die Gegenwart

# Ausgabe einer Zusammenfassung der Ensemble-Modellierung Zukunft
myproj.baseline.ensemble

# Extraktion der Geodaten aus der Projektion
myproj.baseline.ensemble.geo <- get_predictions(myproj.baseline.ensemble)

# Plot der Modellierungsergebnisse
plot(myproj.baseline.ensemble.geo)

# Plot der mittleren vorhergesagten Suitability ueber alle Modellalgorithmen
# "testspecies" ggf. durch den inhalt von "speciesname" ersetzen!
plot(myproj.baseline.ensemble.geo$testspecies_EMmeanByROC_mergedAlgo_mergedRun_mergedData)

# Abspeichern der einzelnen Layer als tif-Dateien zur weiteren Verwendung in anderen Programmen.
# Interessant ist vornehmlich die Datei deren Name "EMmeanByROC_mergedAlgo_mergedRun_mergedData" enthaelt.
writeRaster(x=myproj.baseline.ensemble.geo, filename=paste0(path.output, speciesname, "_ensemble_baseline"), bylayer=TRUE, suffix="names", format="GTiff", overwrite=TRUE)  

#############################################################
# 7. Zukunftsprojektion fuer die einzelnen Modellalgorithmen

# Projektion der Modellierungsergebnisse unter zukuenftigen Umweltbedingungen fuer die einzelnen Modellalgorithmen
# Dies ist komplett analog zu Abschnitt 4, mit dem Unterschied dass andere Umweltdaten (envir.future) verwendet werden
# Zur naeheren Erlaeuterung siehe auch: http://www.inside-r.org/packages/cran/biomod2/docs/BIOMOD_Projection
myproj.future <- BIOMOD_Projection(modeling.output = mymodels, # Modellierungsergebnis (erstelltes Objekt)
                                        new.env = envir.future,             # Set an erklaerenden (Umwelt-)Variablen (hier: Zukunft)
                                        proj.name = 'future',               # Name fuer die Projektion (frei waehlbar;legt einen neuen Ordner mit diesem Namen an)
                                        selected.models = 'all',            # Auswahl an Modellen, die verwendet werden sollen (all: alle Modelle werden verwendet)
                                        binary.meth = 'ROC',                # Methode zur Umrechnung der Projektion in 0/1 Daten (binaere Daten)
                                        compress = F,                       # Komprimierungsmethode beim Abspeichern von Objekten
                                        build.clamping.mask = F,            # Wenn TRUE: Identifiziert Regionen, in denen die Projektionen unsicher sind
                                        output.format = '.img')             # Objekt kann als .RData, .grd oder .img abgespeichert werden

# Zusammenfassung der Zukunfts-Projektion
myproj.future

# Darstellung der Zukunfts-Projektion  der einzelnen Modell-Algorithmen
plot(myproj.future)

# Ausgabe der Projektionen als Multi-layer Geo-Tiff
writeRaster(get_predictions(myproj.future), paste0(path.output, speciesname, "_all_models_future"), format="GTiff")

#############################################################
# 8. Zukunftsprojektion fuer die Ensemble-Modellierung

# Ensemble-Modellierung Zukunft (Berechnung des Ensembles aus den verwendeten Modellalgorithmen)
myproj.future.ensemble <- BIOMOD_EnsembleForecasting(EM.output = mymodels.ensemble,                   # Ergebnis der Ensembleberechnung fuer die Gegenwart
                                         projection.output = myproj.future)   # Projektion des Ensembles in die Zukunft

# Ausgabe einer Zusammenfassung der Ensemble-Modellierung Zukunft
myproj.future.ensemble

# Extraktion der Geodaten aus der Projektion
myproj.future.ensemble.geo <- get_predictions(myproj.future.ensemble)

# Plot der Modellierungsergebnisse
plot(myproj.future.ensemble.geo)

# Plot der mittleren vorhergesagten Suitability ueber alle Modellalgorithmen
# "testspecies" ggf. durch den inhalt von "speciesname" ersetzen!
plot(myproj.future.ensemble.geo$testspecies_EMmeanByROC_mergedAlgo_mergedRun_mergedData)

# Abspeichern der einzelnen Layer als tif-Dateien zur weiteren Verwendung in anderen Programmen.
# Interessant ist vornehmlich die Datei deren Name "EMmeanByROC_mergedAlgo_mergedRun_mergedData" enthaelt.
writeRaster(x=myproj.future.ensemble.geo, filename=paste0(path.output, speciesname, "_ensemble_future"), bylayer=TRUE, suffix="names", format="GTiff", overwrite=TRUE)  


#############################################################
# 9. Abschliessend Erneute Sicherung des gesamten Workspaces
# (kann mit load("XY.RData") wieder in R eingelesen werden)
save.image(paste0(path.data, "workspace-backup_SDM_final", speciesname, ".RData"))



