##########
## Funktion zum Zuschneiden einer Raster*-Objectes auf ein Polygon
rt.clip <- function(data, cropshape){
  dat.crop <- crop(data, cropshape)     # Grobes Zuschneiden auf die Ausdehnung des Polygons
  dat.crop <- mask(dat.crop, cropshape) # Daten ausserhalb des Polygons auf "NA" setzen
  return(dat.crop)
}

##########
## Funktion zur Extraktion des Mittelwertes einer 
## Temperaturvariable über den von einer einzelnen .nc-Datei 
## abgedeckten Zeitabschnittes für einen gegebenen Tag im 
## Jahr. Z.B. Mittelwert für den 1. Januar der Jahre 2005-2010
rt.dailymean <- function(day="01", month="01", startyear, endyear, infile, varname, crs.in, fun=mean){
  # day              = zu extrahierender Tag (01-31); als character (z.B. "01")
  # month            = Monat in dem dieser Tag liegt (01-12); als character (z.B. "01")
  # startyear        = erstes zu beruecksichtigendes Jahr
  # emdyear          = letztes zu beruecksichtigendes Jahr
  # path.infile      = Dateipfad zur .nc-Datei
  # path.outfiles    = Dateipfad zum gewuenschten Ausgabeordner
  # varname          = Name der Klimavariablen in der netcdf-Datei
  # fun              = wenn andere Funtion als der Mittelwert
  #                    gewünscht is, hier eine von "calc"
  #                    (Paket "raster") akzeptierte Funktion
  #                    angeben
  ########################################
  # netcdf-datei als RasterStack einlesen
  suppressWarnings(dat <- stack(infile, varname=varname))
  # wenn gewuenscht, Projektionsinformationen hinzufuegen
  if(!missing(crs.in)){
    crs(dat) <- crs.in
  }
  # Daten für den gewuenschten Tag und alle Jahre auswaehlen
  years <- startyear:endyear
  alldates <- dat@z[[1]]
  seldates <- which(substr(alldates, start=6, stop=10) == paste(month, day, sep="-") & substr(alldates, start=1, stop=4) %in% as.character(years))
  
  dat.stack <- stack(dat[[seldates]])
  # Mittelwert über alle Jahre bilden
  dat.mean <- calc(dat.stack, fun=fun)
  return(dat.mean)
}

##########
## Klimadaten-Dateien umfassen oft nur wenige Jahre pro netcdf-Datei
## Diese Funktion ruft "rt.dailymean" für mehrere dieser Dateien auf und
## aggregiert die daraus resultierenden Ausgaben in eine einzige Datei.
rt.aggregate_timesteps <- function(infiles, fun=mean, ...){
  # infiles          = character-Vektor mit Dateipfaden der Eingabedateien
  # fun              = wenn andere Funtion als der Mittelwert
  #                    gewünscht is, hier eine von "calc"
  #                    (Paket "raster") akzeptierte Funktion
  #                    angeben
  # ...              = Funtionsparameter, die an "rt.dailymean" weitergereicht weren
  ########################################
  # leeren RasterStack erzeugen
  dat.stack <- stack()
  # "rt.dailymean" für alle Eingabedateien aufrufen und 
  # Ausgabe an RasterStack anhängen
  for(infile in infiles){
    dat.temp <- rt.dailymean(infile=infile, ...)
    dat.stack <- stack(dat.stack, dat.temp)
  }
  # Mittelwert über alle Eingabezeitsbschnitte bilden
  dat.stack.mean <- calc(dat.stack, fun=fun)
  return(dat.stack.mean)
}

##########
## "rt.dailymean" bzw. "rt.aggregate_timesteps" behandeln jeweils nur einen Tag des Jahres
## (z.B. 1. Januar). Diese Funktion ruft "rt.dailymean" bzw. "rt.aggregate_timesteps"
## für alle 365 Tage im Jahr auf (ignoriert 29. Februar)
rt.extract365 <- function(path.outfiles, infiles, fname.prefix="", cropshape, cropmode="clip", overwrite=FALSE, ...){
  # path.outfiles    = Dateipfad zum gewünschten Ausgabeordner
  # infiles          = character-Vektor mit Dateipfaden der Eingabedateien
  # fname.prefix     = optionaler Dateinamensvorsatz fuer die Ausgabedateien
  # cropshape        = optional: SpatialPolygons-Objekt für Zuschnitt
  # cropmode         = "clip": Feiner Zuschnitt entlang der Umrisse des Polygons
  #                    "crop": Grober Zuschnitt auf die Ausmasse des Polygons
  # overwrite        = sollen bereits existierende Dateien gleichen Namens
  #                    überschrieben werden?
  # ...              = Funtionsparameter, die an "rt.dailymean" bzw. "rt.aggregate_timesteps"
  #                    weitergereicht weren
  ########################################
  # überprüfen ob Ausgabe-Verzeichnis bereits vorhanden,
  # sonst anlegen
  if(!file.exists(path.outfiles)){
    message("creating output directory at", path.outfiles)
    dir.create(path.outfiles)
  }
  # erzeuge eine Auflistung aller Tage im Jahr
  dates <- seq.Date(from=as.Date("2010-01-01"), to=as.Date("2010-12-31"), by=1)
  months <- substr(dates, start=6, stop=7)
  days <- substr(dates, start=9, stop=10)
  # Datenextraktion
  for(i in 1:365){
    message("extracting day ", days[i], " of month ", months[i], " …")
    # wenn nur eine Eingabedatei angegeben, rufe "rt.dailymean" direkt auf ...
    if(length(infiles)==1){
      dat <- rt.dailymean(day=days[i], month=months[i], infile=infiles, ...)
      # ... ansonsten benutze "rt.aggregate_timesteps"
    }else{
      dat <- rt.aggregate_timesteps(day=days[i], month=months[i], infiles=infiles, ...)
    }
    # schreibe Datei ins Ausgabeverzeichnis
    if(!missing(cropshape)){
      if(cropmode=="clip"){
        dat <- rt.clip(data=dat, cropshape=cropshape)
      }else{
        if(cropmode=="crop"){
          dat <- crop(dat, cropshape)
        }
      }
    }
    writeRaster(dat, filename=paste0(path.outfiles, fname.prefix, "_", months[i], "-", days[i], ".tif"), format="GTiff", overwrite=overwrite, NAflag=-9999)
  }
}

##########
## Anzahl an Tagen ueber unter einem Grenzwert
rt.threshold <- function(filenames, threshold, thresholdside="above"){
  if(missing(thresholdside)){
    stop('Argument "thresholdside" missing. Please specify whether you are interested in values above or below the threshold by entering thresholdside="above" or thresholdside=below"')
  }
  dat <- stack(filenames)
  if(thresholdside == "above"){
    m <- c(-Inf, threshold, 0, threshold, Inf, 1)
    rcl <- matrix(m, ncol=3, byrow=TRUE)
    dat.rcl <- reclassify(dat, rcl=rcl)
  }else{
    m <- c(-Inf, threshold, 1, threshold, Inf, 0)
    rcl <- matrix(m, ncol=3, byrow=TRUE)
    dat.rcl <- reclassify(dat, rcl=rcl)
  }
  dat <- sum(dat.rcl) # Aufsummieren der Anzahl an Tagen ueber/unter dem Grenzwert
  return(dat)
}

##########
## Einlesen von Daten fuer rt.consdays
rt.consdaysdata <- function(infiles){
  # infiles: Ein Vektor der Dateinamen zu Rasterdateien beinhaltet
  dat <- stack(infiles)                  # Eingangsdaten als Raster Stack einladen
  dat <- as(dat, "SpatialGridDataFrame") # in SpatialGridDataFrame umwandeln
  return(dat)
}

##########
## Anzahl aufeinanderfolgender Tage >= eines gegebenen (Temperatur-) Wertes
rt.consdays <- function(data, referencevalue, startday=1, endday=365){
  # data           = multi-layer SpatialGridDataFrame erzeugt durch rt.consdaysdata
  # referencevalue = Finde maximale Anzahl aufeinanderfolgender Tage bei/ueber diesem Grenzwert (Temperatur)
  # startday       = erster Tag im Jahr der beruecksichtigt werden soll
  #                  (numerisch, 1-364)
  # endday         = letzter Tag im Jahr der beruecksichtigt werden soll
  #                  (numerisch, 2-365)
  
  # Rohdaten ohne Geoinformation extrahieren
  data.raw <- data@data
  # Leeren Vektor fuer die Anzahl an Tagen bei/ueber Grenzwert anlegen
  days.number <- c()
  pb <- txtProgressBar(min = 0, max = nrow(data.raw), style = 3) # Fortschrittsbalken intialisieren
  # Wiederhole fuer jede Rasterzelle:
  for(thispixel in 1:nrow(data.raw)){
    thispixel.values <- data.raw[thispixel, ] # Daten fuer diese Rasterzelle aus Gesamtdatensatz herausziehen
    if(is.na(thispixel.values[1])){           # Wenn diese Rasterzelle keine Daten enthaelt, ueberspringe ihn
      j <- NA
    }else{
      i <- 0                                  # Zaehler fuer die Anzahl aufeinanderfolgender Tage
      j <- 0                                  # zum speichern des maximalwertes von i
      # Wiederhole fuer jeden Tag im Jahr:
      for(thisday in startday:endday){        
        if(thispixel.values[thisday] >= referencevalue){  # Wenn der Wert fuer diesen Tag groesser ist als die Referenztemperatur...
          i <- i + 1                          # erhoehe i um 1
        }else{                                # wenn nicht beende Zaehlung und ...
          if(i > j){j <- i}                   # speichere i in j wenn dies die groesste bisher gefundene Zahl an aufeinanderfolgenden Tagen ist
          i <- 0                              # setze i zurueck auf 0 und fahre fort
        }
        if(i > j){j <- i} # fall ueber den gesamten Zeitraum ununterbrochen >= referencevalue, muss ganz am Ende i manuell in j ueberfuerht werden
      }
    }
    days.number <- c(days.number, j)          # Wenn Operation fuer Rasterzelle beendet, j an days.number anhaengen
    setTxtProgressBar(pb, thispixel)          # Fortschrittsbalken aktualisieren
  }
  close(pb) # Fortschrittsbalken beenden
  return(days.number) # Ausgabe
}

##########
## kuerzest moegliche EIP
rt.eip_duration <- function(data, temperatures, eips, startday=1, endday=365){
  # data         = multi-layer SpatialGridDataFrame erzeugt durch rt.consdaysdata
  # temperatures = Vektor mit Referenztemperaturen
  # eips         = Vektor mit den den Referenztemperaturen entsprechenden EIP
  # startday     = erster Tag im Jahr der beruecksichtigt werden soll
  #                (numerisch, 1-364)
  # endday       = letzter Tag im Jahr der beruecksichtigt werden soll
  #                (numerisch, 2-365)
  eip <- rep(Inf, times=nrow(data@data)) # Vektor mit Laenge = Anzahl der Rasterzellen erzeugen; Werte = Unendlich
  # Fuer jede angegebene EIP:
  for(i in 1:length(eips)){
    message("running rt.consdays for EIP of ", eips[i], " days …")
    # Maximale Anzahl aufeinanderfolgender Tage bei der dieser EIP entsprechenden Temperatur
    consdays <- rt.consdays(data=dat, referencevalue=temperatures[i], startday=startday, endday=endday)
    # Wenn Anzahl der Tage bei dieser Temperatur, EIP fuer diese Rasterzelle eintragen, ansonsten Unendlich
    eip.temp <- consdays
    eip.temp[which(consdays<eips[i])] <- Inf
    eip.temp[which(consdays>=eips[i])] <- eips[i]
    # Wenn die in diesem Durchlauf gefundene EIP fuer eine Rasterzelle kuerzer ist als
    # eine zuvor gefundene, letztere ueberschreiben
    eip <- pmin(eip, eip.temp)
  }
  # Vektor in Geodaten umwandeln und ausgeben
  eip.geo <- data
  eip.geo@data <- data.frame(EIP=eip)
  eip.geo <- raster(eip.geo)
  return(eip.geo)
}

##########
## EIP transformieren und bei Bedarf an Aufloesung eines anderen Rasters anpassen
rt.transformEIP <- function(eip, refraster){
  # Kehrwert bilden
  eip.norm <- 1/eip
  # Auf werte zwischen 0 und 1 normieren
  eip.norm[] <- decostand(eip.norm[], method="range", na.rm=T)
  # wenn gewuenscht an Aufloesung eines anderen Rasters anpassen
  if(!missing(refraster)){
    eip.norm <- resample(eip.norm, refraster, method="ngb")
  }
  return(eip.norm)
}

